"""Unflakable API."""

#  Copyright (c) 2022 Developer Innovations, LLC

from typing import List, Optional, TYPE_CHECKING

import sys
import logging
import pkg_resources
import platform
import pprint
import requests

if TYPE_CHECKING:
    from typing_extensions import NotRequired, TypedDict
else:
    from typing import Dict

    NotRequired = Optional
    TypedDict = Dict

BASE_URL = 'https://app.unflakable.com'
TEST_NAME_ENTRY_MAX_LENGTH = 4096
PACKAGE_VERSION = pkg_resources.get_distribution('pytest_unflakable').version
PLATFORM_VERSION = platform.platform()
PYTEST_VERSION = pkg_resources.get_distribution('pytest').version
PYTHON_VERSION = f'{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}'
USER_AGENT = (
    f'unflakable-pytest-plugin/{PACKAGE_VERSION} (PyTest {PYTEST_VERSION}; '
    f'Python {PYTHON_VERSION}; Platform {PLATFORM_VERSION})'
)


class TestRef(TypedDict):
    test_id: str
    filename: str
    name: List[str]


class TestSuiteManifest(TypedDict):
    quarantined_tests: List[TestRef]


if TYPE_CHECKING:
    from typing import Literal

    TestAttemptResult = Literal['pass', 'fail', 'quarantined']
else:
    TestAttemptResult = str


class TestRunAttemptRecord(TypedDict):
    start_time: NotRequired[Optional[str]]
    end_time: NotRequired[Optional[str]]
    duration_ms: NotRequired[Optional[int]]
    result: TestAttemptResult


class TestRunRecord(TypedDict):
    filename: str
    name: List[str]
    attempts: List[TestRunAttemptRecord]


class CreateTestSuiteRunRequest(TypedDict):
    branch: NotRequired[Optional[str]]
    commit: NotRequired[Optional[str]]
    start_time: str
    end_time: str
    test_runs: List[TestRunRecord]


class TestSuiteRunSummary(TypedDict):
    run_id: str
    suite_id: str
    branch: NotRequired[Optional[str]]
    commit: NotRequired[Optional[str]]
    start_time: str
    end_time: str
    num_tests: int
    num_pass: int
    num_fail: int
    num_flake: int
    num_quarantined: int


def create_test_suite_run(
        request: CreateTestSuiteRunRequest,
        test_suite_id: str,
        api_key: str,
        base_url: Optional[str],
        insecure_disable_tls_validation: bool,
        logger: logging.Logger,
) -> TestSuiteRunSummary:
    logger.debug(f'creating test suite run {pprint.pformat(request)}')

    run_response = requests.post(
        url=(
            f'{base_url if base_url is not None else BASE_URL}/api/v1/test-suites/{test_suite_id}'
            '/runs'
        ),
        headers={
            'Authorization': f'Bearer {api_key}',
            'Content-Type': 'application/json',
            'User-Agent': USER_AGENT,
        },
        json=request,
        verify=not insecure_disable_tls_validation,
    )
    run_response.raise_for_status()

    summary: TestSuiteRunSummary = run_response.json()
    logger.debug(f'received response: {pprint.pformat(summary)}')

    return summary


def get_test_suite_manifest(
        test_suite_id: str,
        api_key: str,
        base_url: Optional[str],
        insecure_disable_tls_validation: bool,
        logger: logging.Logger,
) -> TestSuiteManifest:
    logger.debug(f'fetching manifest for test suite {test_suite_id}')

    manifest_response = requests.get(
        url=(
            f'{base_url if base_url is not None else BASE_URL}/api/v1/test-suites/{test_suite_id}'
            '/manifest'
        ),
        headers={
            'Authorization': f'Bearer {api_key}',
            'User-Agent': USER_AGENT,
        },
        verify=not insecure_disable_tls_validation,
    )
    manifest_response.raise_for_status()

    manifest: TestSuiteManifest = manifest_response.json()
    logger.debug(f'received response: {pprint.pformat(manifest)}')

    return manifest


def build_test_suite_run_url(
        test_suite_id: str,
        run_id: str,
        base_url: Optional[str]
) -> str:
    return (
        f'{base_url if base_url is not None else BASE_URL}/test-suites/{test_suite_id}'
        f'/runs/{run_id}'
    )



#  Copyright (c) 2022 Developer Innovations, LLC
import logging
from typing import Optional, List
from subprocess import run
import sys

GIT_ERROR_STR = 'WARNING: Unflakable failed to auto-detect current git branch and commit'
GIT_ERROR_HINT = (
    'HINT: pass the --branch and --commit command-line arguments or disable git '
    'auto-detection by passing --no-git-auto-detect'
)


def _run_git_command(args: List[str]) -> Optional[str]:
    try:
        git_outcome = run(
            ['git'] + args,
            capture_output=True,
            text=True,
        )
    except Exception as e:
        sys.stderr.write(f'{GIT_ERROR_STR}: {repr(e)}\n{GIT_ERROR_HINT}\n')
        return None

    if git_outcome is None:
        return None
    elif git_outcome.returncode != 0:
        sys.stderr.write(
            f'{GIT_ERROR_STR}: git exited with non-zero exit code {git_outcome.returncode}: '
            f'{git_outcome.stderr}\n{GIT_ERROR_HINT}\n'
        )
        return None
    else:
        return git_outcome.stdout


def get_current_git_branch(commit_sha: str, logger: logging.Logger) -> Optional[str]:
    # In the common case (an attached HEAD), we can just use `git rev-parse`.
    head_ref_raw = _run_git_command(['rev-parse', '--abbrev-ref', 'HEAD'])
    head_ref = head_ref_raw if head_ref_raw is None else head_ref_raw.rstrip()

    # If `git rev-parse` returns `HEAD`, then we have a detached head, and we need to see if the
    # current commit SHA matches any known refs (i.e., local/remote branches or tags). This happens
    # when running GitHub Actions in response to a `pull_request` event. In that case, the commit
    # is a detached HEAD, but there's a `refs/remotes/pull/PR_NUMBER/merge` ref we can use as the
    # "branch" (abbreviated to pull/PR_NUMBER/merge).
    if head_ref != 'HEAD':
        return head_ref

    # The code below runs the equivalent of `git show-ref | grep $(git rev-parse HEAD)`.
    git_output = _run_git_command(['show-ref'])
    if git_output is None:
        return git_output

    matching_refs = [
        line.split(' ', 1)[1] for line in git_output.splitlines()
        if line.startswith(commit_sha + ' ')
    ]
    logger.debug(
        'git show-ref returned %d ref(s) for SHA %s: %s',
        len(matching_refs),
        commit_sha,
        ', '.join(matching_refs)
    )

    if len(matching_refs) == 0:
        return None

    # `git show-ref` returns refs sorted lexicographically:
    #   refs/heads/*
    #   refs/remotes/*
    #   refs/stash
    #   refs/tags/*
    # We just take the first matching ref and use its abbreviation (i.e., removing the refs/remotes
    # prefix) as the branch name. Users can override this behavior by setting the UNFLAKABLE_BRANCH
    # environment variable.
    abbreviated_ref = _run_git_command(['rev-parse', '--abbrev-ref', matching_refs[0]])
    if abbreviated_ref is None:
        return None
    else:
        return abbreviated_ref.rstrip()


def get_current_git_commit() -> Optional[str]:
    git_commit = _run_git_command(['rev-parse', 'HEAD'])
    if git_commit is None:
        return None
    else:
        return git_commit.rstrip()



"""Plugin implementation."""
#  Copyright (c) 2022 Developer Innovations, LLC

from enum import Enum
from pathlib import Path
from typing import (
    Any, Dict, List, Union, Tuple, Mapping, Optional, cast, Generator, Set, TYPE_CHECKING
)

import logging
import pprint

import py
import pytest
import _pytest
import sys
from time import time
from datetime import datetime, timezone

from ._api import (
    get_test_suite_manifest,
    create_test_suite_run,
    build_test_suite_run_url,
    CreateTestSuiteRunRequest,
    TestRunAttemptRecord,
    TestRunRecord, TestAttemptResult, TestSuiteManifest,
)

if TYPE_CHECKING:
    from typing import Literal

    # Most of these types aren't defined or exported in older versions of pytest, but we only need
    # the type checking to work on newer versions.
    CallInfo = pytest.CallInfo[None]
    CallPhase = Literal['setup', 'call', 'teardown']
    CollectReport = pytest.CollectReport
    Config = pytest.Config
else:
    CallInfo = object
    CallPhase = str
    CollectReport = object
    Config = object

TestPathAndName = Tuple[str, Tuple[str, ...]]
QuarantinedTests = Set[TestPathAndName]
PathCompat = Union[py.path.local, Path]


class QuarantineMode(Enum):
    IGNORE_FAILURES = 1
    NO_QUARANTINE = 2
    SKIP_TESTS = 3


def _ts_to_rfc3339(ts_sec: float) -> str:
    return datetime.fromtimestamp(ts_sec, timezone.utc).isoformat('T', 'microseconds')


# Support both pre-7.0 and post-7.0 versions of PyTest.
def node_path(session: Union[pytest.Item, pytest.Session]) -> PathCompat:
    if hasattr(session, 'path'):
        return session.path
    else:
        return session.fspath


def relative_to(path: PathCompat, base: PathCompat) -> str:
    if hasattr(path, 'relative_to'):
        return str(path.relative_to(base))  # type: ignore
    else:
        return str(path.relto(base))  # type: ignore


def item_name(item: _pytest.nodes.Node) -> Tuple[str, ...]:
    if isinstance(item, pytest.Module):
        return tuple()
    elif item.parent is None:
        return (item.name,)
    elif isinstance(item, pytest.Class) or isinstance(item, pytest.Function):
        return item_name(item.parent) + (item.name,)
    else:
        # Prior to PyTest 7.0, there was a node in between Class and Function called Instance. Its
        # name is `()`, which doesn't add any value, so we just filter it out.
        return item_name(item.parent)


class UnflakableReport(_pytest.reports.BaseReport):
    unflakable_attempt: int
    unflakable_is_quarantined: bool
    unflakable_prior_non_teardown_failures: int
    unflakable_start_time: float
    unflakable_end_time: float
    unflakable_test_name: Tuple[str, ...]
    unflakable_filename: str

    # The PyTest TerminalReporter emits error messages for `failed` and `error` categories, but
    # not for unknown categories like `flaky` and `quarantined`. To ensure that errors get logged
    # properly, we generate two reports for flaky/quarantined test:
    #   1. One for logging with category `failed` or `error` and count_towards_summary() returning
    #      False. The TerminalReporter will log the corresponding errors, but this report will
    #      not contribute toward the summary stats at the end. For this one,
    #      `unflakable_fake_report_for_logging` is True.
    #   2. One with category `flaky`/`quarantined` that contributes to the summary stats but
    #      doesn't trigger any additional logging of errors. For this one,
    #      `unflakable_fake_report_for_logging` is False.
    unflakable_fake_report_for_logging: bool

    def __init__(
            self,
            unflakable_attempt: int,
            unflakable_is_quarantined: bool,
            unflakable_prior_non_teardown_failures: int,
            unflakable_start_time: float,
            unflakable_end_time: float,
            unflakable_test_name: Tuple[str, ...],
            unflakable_filename: str,
            unflakable_fake_report_for_logging: bool = False,
            **kw: Any
    ) -> None:
        super().__init__(
            unflakable_attempt=unflakable_attempt,
            unflakable_is_quarantined=unflakable_is_quarantined,
            unflakable_prior_non_teardown_failures=unflakable_prior_non_teardown_failures,
            unflakable_start_time=unflakable_start_time,
            unflakable_end_time=unflakable_end_time,
            unflakable_test_name=unflakable_test_name,
            unflakable_filename=unflakable_filename,
            unflakable_fake_report_for_logging=unflakable_fake_report_for_logging,
            **kw,
        )

    def __repr__(self) -> str:
        return '<UnflakableReport {!r} when={!r} outcome={!r}>'.format(
            self.nodeid, self.when, self.outcome
        )

    def to_fake_for_logging(self) -> 'UnflakableReport':
        assert not self.unflakable_fake_report_for_logging
        copy = UnflakableReport(**self.__dict__)
        copy.unflakable_fake_report_for_logging = True
        return copy

    @property
    def count_towards_summary(self) -> bool:
        return not self.unflakable_fake_report_for_logging
        # if self.failed and self.unflakable_is_quarantined and self.unflakable_attempt > 0:
        #     return False
        # return True


ItemReports = Dict[CallPhase, UnflakableReport]


class UnflakablePlugin:
    api_key: str
    base_url: Optional[str]
    branch: Optional[str]
    commit: Optional[str]
    failure_retries: int
    insecure_disable_tls_validation: bool
    quarantine_mode: QuarantineMode
    test_suite_id: str
    upload_results: bool
    logger: logging.Logger

    is_xdist_worker: bool
    item_reports: Dict[TestPathAndName, List[ItemReports]]
    manifest: Optional[TestSuiteManifest]
    quarantined_tests: Set[TestPathAndName]
    non_teardown_failures: Set[str]
    session: Optional[pytest.Session]
    num_tests_quarantined: int
    end_time: float
    start_time: float

    def __init__(
            self,
            api_key: str,
            base_url: Optional[str],
            branch: Optional[str],
            commit: Optional[str],
            failure_retries: int,
            insecure_disable_tls_validation: bool,
            quarantine_mode: QuarantineMode,
            test_suite_id: str,
            upload_results: bool,
            logger: logging.Logger,
            worker_manifest: Optional[TestSuiteManifest],
            is_xdist_worker: bool,
    ):
        self.api_key = api_key
        self.base_url = base_url
        self.branch = branch
        self.commit = commit
        self.failure_retries = failure_retries
        self.insecure_disable_tls_validation = insecure_disable_tls_validation
        self.quarantine_mode = quarantine_mode
        self.test_suite_id = test_suite_id
        self.upload_results = upload_results
        self.logger = logger

        self.is_xdist_worker = is_xdist_worker
        self.item_reports = {}

        self.manifest = worker_manifest
        if self.manifest is None:
            try:
                self.manifest = get_test_suite_manifest(
                    test_suite_id=self.test_suite_id,
                    api_key=self.api_key,
                    base_url=self.base_url,
                    insecure_disable_tls_validation=self.insecure_disable_tls_validation,
                    logger=self.logger,
                )
            # IOError is the base class for `requests.RequestException`.
            except IOError as e:
                sys.stderr.write(
                    ('ERROR: Failed to get Unflakable manifest: %s\nTest failures will NOT be'
                     ' quarantined.\n') % (repr(e))),
        else:
            logger.debug(
                f'xdist worker received manifest for test suite {self.test_suite_id}: '
                f'{pprint.pformat(self.manifest)}'
            )

        self.quarantined_tests = set([
            (quarantined_test['filename'], tuple(quarantined_test['name'])) for
            quarantined_test
            in self.manifest['quarantined_tests']
        ]) if self.manifest is not None else set()

        self.non_teardown_failures = set()
        self.session = None
        self.num_tests_quarantined = 0
        self.end_time = 0.
        self.start_time = 0.

    # Lets us filter out quarantined tests when `quarantine_mode` is `skip_tests`.
    def pytest_collection_modifyitems(
            self,
            session: pytest.Session,
            config: Config,
            items: List[pytest.Item],
    ) -> None:
        self.logger.debug('called hook pytest_collection_modifyitems')
        for idx, item in enumerate(items):
            test_path = relative_to(node_path(item), node_path(session))
            test_name = item_name(item)
            if (test_path, test_name) in self.quarantined_tests:
                self.logger.info(
                    'test `%s` in file %s is quarantined', '.'.join(test_name), test_path
                )
                if self.quarantine_mode == QuarantineMode.SKIP_TESTS:
                    item.add_marker(pytest.mark.skip)
            else:
                self.logger.debug('test `%s` in file %s is not quarantined', test_name, test_path)

    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_protocol(self, item: pytest.Item, nextitem: Optional[pytest.Item]) -> bool:
        self.logger.debug(f'called hook pytest_runtest_protocol: {item.nodeid}')
        ihook = item.ihook
        ihook.pytest_runtest_logstart(nodeid=item.nodeid, location=item.location)

        assert self.session
        test_filename = relative_to(node_path(item), node_path(self.session))
        test_name = item_name(item)
        for attempt in range(item.config.option.unflakable_failure_retries + 1):
            if attempt > 0:
                self.logger.info(
                    'retrying test `%s` in file %s (attempt %d of %d)',
                    '.'.join(test_name),
                    relative_to(node_path(item), node_path(item.session)),
                    attempt + 1,
                    item.config.option.unflakable_failure_retries + 1,
                )

            # It's not ideal to use this protected function directly, but it's the only way to
            # avoid calling the pytest_runtest_logstart() and pytest_runtest_logfinish() hooks
            # multiple times for retried tests, which somewhat garbles the output. This is
            # similar to the approach that pytest_rerunfailures takes:
            # https://github.com/pytest-dev/pytest-rerunfailures/blob/e80c12eb9456b7646c3c6610b2c08420e9247cd4/pytest_rerunfailures.py#L525
            _pytest.runner.runtestprotocol(item, nextitem=nextitem)

            item_reports = self.item_reports.get((test_filename, test_name), [])
            if all([item_report.passed or item_report.skipped for item_report in
                    item_reports[-1].values()]):
                break

        ihook.pytest_runtest_logfinish(nodeid=item.nodeid, location=item.location)
        return True

    def pytest_runtest_makereport(
            self,
            item: pytest.Item,
            call: CallInfo,
    ) -> UnflakableReport:
        # Older versions of PyTest don't export this from the main pytest package.
        report = _pytest.reports.TestReport.from_item_and_call(item, call)

        # NB: For xdist, this hook runs in the worker.
        self.logger.debug(
            f'called hook pytest_runtest_makereport: {item.nodeid} ({call.when}) - {report.outcome}'
        )

        if item.parent:
            pass

        test_filename = relative_to(node_path(item), node_path(item.session))
        test_name = item_name(item)
        is_quarantined = (test_filename, test_name) in self.quarantined_tests and (
                self.quarantine_mode == QuarantineMode.IGNORE_FAILURES)

        assert self.session

        item_reports = self.item_reports.get((test_filename, test_name), [])
        if len(item_reports) == 0:
            attempt = 0
        elif call.when in item_reports[-1]:
            attempt = len(item_reports)
        else:
            attempt = len(item_reports) - 1

        # NB: When running with xdist, this function only gets called on the workers, so we don't
        # update any plugin state until pytest_report_teststatus() gets called on the controller.

        unflakable_report = UnflakableReport(
            unflakable_attempt=attempt,
            unflakable_is_quarantined=is_quarantined,
            unflakable_prior_non_teardown_failures=item.nodeid in self.non_teardown_failures,
            unflakable_start_time=call.start,
            unflakable_end_time=call.stop,
            unflakable_test_name=test_name,
            unflakable_filename=test_filename,
            **report.__dict__,
        )
        return unflakable_report

    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_logreport(
            self,
            report: UnflakableReport,
    ) -> None:
        # NB: This gets called on the xdist worker, which then proxies it to the controller and
        # calls it again there.
        self.logger.debug(
            f'called hook pytest_runtest_logreport {report.nodeid} ({report.when}) - '
            f'{report.outcome}'
        )

        # Recursive base case.
        if report.unflakable_fake_report_for_logging:
            return None

        # NB: We do this here and not in pytest_runtest_makereport() because the PyTest `skipping`
        # module has a pytest_runtest_makereport() wrapper hook for handling xfail strict and other
        # cases that modify the report. We need to act on the final version of the report after
        # those modifications.
        if report.failed and report.when != 'teardown':
            self.non_teardown_failures.add(report.nodeid)

        item_reports: List[ItemReports] = (
            self.item_reports.setdefault(
                (report.unflakable_filename, report.unflakable_test_name),
                [],
            )
        )
        if report.when == 'setup':
            item_reports.append({'setup': report})
        elif report.when in ['call', 'teardown'] and len(item_reports) > 0:
            item_reports[-1][cast(CallPhase, report.when)] = report

        if report.failed and not self.is_xdist_worker:
            assert self.session

            if report.unflakable_is_quarantined and (
                    self.quarantine_mode == QuarantineMode.IGNORE_FAILURES
            ):
                self.num_tests_quarantined += 1
                self.session.ihook.pytest_runtest_logreport(report=report.to_fake_for_logging())
            # If the last attempt fails, it'll get logged as a failure/error anyway, so we don't
            # need to create an extra report.
            elif report.unflakable_attempt < self.failure_retries:
                self.session.ihook.pytest_runtest_logreport(report=report.to_fake_for_logging())

    def pytest_report_teststatus(
            self,
            report: Union[CollectReport, UnflakableReport],
            config: Config,
    ) -> Optional[Tuple[str, str, Union[str, Tuple[str, Mapping[str, bool]]]]]:
        self.logger.debug(
            f'called hook pytest_report_teststatus: {report.nodeid} ({report.when}) - '
            f'{report.outcome}'
        )

        item_reports = self.item_reports.get(
            (report.unflakable_filename, report.unflakable_test_name), [])
        attempt = max(len(item_reports) - 1, 0)

        if report.failed:
            if report.unflakable_is_quarantined and (
                    self.quarantine_mode == QuarantineMode.IGNORE_FAILURES
            ):
                if report.unflakable_fake_report_for_logging:
                    return 'error' if report.when in ['setup', 'teardown'] else 'failed', '', ''
                return (
                    # Don't double-count quarantined tests; we already returned `quarantined` for
                    # the category during the first attempt, so return an empty string for retries
                    # to prevent the final stats from double-counting multiple runs of the same
                    # test.
                    'quarantined' if attempt == 0 else '',
                    'Q',
                    (
                        'ERROR (quarantined)' if attempt == 0 else 'ERROR (retry, quarantined)',
                        {'purple': True},
                    ) if report.when in ['setup', 'teardown'] else (
                        'FAILED (quarantined)' if attempt == 0 else 'FAILED (retry, quarantined)',
                        {'purple': True},
                    ),
                )
            elif report.when in ['setup', 'teardown']:
                if report.unflakable_fake_report_for_logging:
                    return 'error', '', ''
                return (
                    # If it's not the last retry attempt, we don't know yet whether the final
                    # outcome is failed or flaky, so we don't return a category.
                    'error' if attempt == self.failure_retries else '',
                    'E' if attempt == 0 else 'R',
                    'ERROR' if attempt == 0 else 'ERROR (retry)',
                )
            elif report.unflakable_fake_report_for_logging:
                return 'failed', '', ''
            else:
                return (
                    # If it's not the last retry attempt, we don't know yet whether the final
                    # outcome is failed or flaky, so we don't return a category.
                    'failed' if attempt == self.failure_retries else '',
                    'F' if attempt == 0 else 'R',
                    'FAILED' if attempt == 0 else 'FAILED (retry)',
                )
        elif (report.passed or (report.skipped and 'xfail' in report.keywords)) and (
                report.when == 'call' and attempt > 0):
            return (
                # Retry passed, so the test is flaky.
                (
                    # We returned the 'quarantined' category on the first failure, so return an
                    # empty string here to avoid double-counting.
                    '' if (report.unflakable_is_quarantined
                           and self.quarantine_mode == QuarantineMode.IGNORE_FAILURES) or (
                              # If only the `teardown` phase has failed in the past, then don't
                              # treat the test as flaky just because the `call` phase passed.
                              not report.unflakable_prior_non_teardown_failures)
                    else 'flaky'
                ),
                'R',
                ('PASSED (retry)', {'green': True}) if report.passed else (
                    'XFAIL (retry)', {'yellow': True}),
            )
        # There's an edge case where just the teardown is flaky, which we don't know until the
        # teardown phase passes on retry without a prior non-teardown failure.
        elif report.passed and report.when == 'teardown' and attempt > 0 and (
                not report.unflakable_is_quarantined or
                self.quarantine_mode != QuarantineMode.IGNORE_FAILURES
        ) and not report.unflakable_prior_non_teardown_failures:
            return (
                'flaky',
                '',
                '',
            )

        return None

    @pytest.hookimpl(tryfirst=True)
    def pytest_report_to_serializable(
            self,
            report: Union[CollectReport, UnflakableReport]
    ) -> Optional[Dict[str, Any]]:
        if isinstance(report, UnflakableReport):
            data = report._to_json()
            data['$report_type'] = report.__class__.__name__
            return data
        return None

    @pytest.hookimpl(tryfirst=True)
    def pytest_report_from_serializable(
            self,
            data: Dict[str, Any],
    ) -> Optional[Union[CollectReport, UnflakableReport]]:
        if data.get('$report_type') == 'UnflakableReport':
            return UnflakableReport._from_json(data)
        return None

    def pytest_sessionstart(self, session: pytest.Session) -> None:
        self.logger.debug('called hook pytest_sessionstart')
        self.session = session
        self.start_time = time()

    # This is a `xdist.workermanage.WorkerController`, but pytest-xdist doesn't provide types.
    def pytest_configure_node(self, node: Any) -> None:
        """
        Hook called by pytest-xdist to configure each worker node.

        We leverage this hook to send the manifest to the worker.
        """
        nodeid = node.workerinput['workerid']
        self.logger.debug(f'called hook pytest_configure_node: {nodeid}')
        if self.manifest is not None:
            node.workerinput['unflakable_manifest'] = self.manifest

    def _build_test_suite_run_request(
            self,
            session: pytest.Session,
    ) -> CreateTestSuiteRunRequest:
        test_runs: List[TestRunRecord] = []
        for (test_filename, test_name), item_reports in self.item_reports.items():
            is_quarantined = (test_filename, test_name) in self.quarantined_tests
            item_attempts: List[TestRunAttemptRecord] = []

            for item_attempt_reports in item_reports:
                # Don't report skipped tests.
                if (
                        'setup' in item_attempt_reports
                        and item_attempt_reports['setup'].skipped
                        and ('call' not in item_attempt_reports
                             or item_attempt_reports['call'].skipped)
                        and 'teardown' in item_attempt_reports
                        and item_attempt_reports['teardown'].passed
                ):
                    continue

                setup_report = item_attempt_reports.get('setup')
                call_report = item_attempt_reports.get('call')
                teardown_report = item_attempt_reports.get('teardown')

                if teardown_report is not None:
                    end_time = teardown_report.unflakable_end_time
                elif call_report is not None:
                    end_time = call_report.unflakable_end_time
                elif setup_report is not None:
                    end_time = setup_report.unflakable_end_time
                else:
                    end_time = None

                if ('setup' in item_attempt_reports and item_attempt_reports['setup'].failed) or (
                        'call' in item_attempt_reports and item_attempt_reports['call'].failed or
                        'teardown' in item_attempt_reports and item_attempt_reports[
                            'teardown'].failed
                ):
                    attempt_result = (
                        cast(TestAttemptResult, 'quarantined')
                        if is_quarantined and (
                                self.quarantine_mode == QuarantineMode.IGNORE_FAILURES) else cast(
                            TestAttemptResult, 'fail')
                    )
                else:
                    attempt_result = cast(TestAttemptResult, 'pass')

                attempt: TestRunAttemptRecord = {
                    'start_time': (
                        _ts_to_rfc3339(setup_report.unflakable_start_time)
                        if setup_report is not None else None
                    ),
                    'end_time': _ts_to_rfc3339(end_time) if end_time is not None else None,
                    'duration_ms': int(
                        (
                                (
                                    item_attempt_reports['setup'].duration
                                    if 'setup' in item_attempt_reports else 0.
                                ) + (
                                    item_attempt_reports['call'].duration
                                    if 'call' in item_attempt_reports else 0.
                                ) + (
                                    item_attempt_reports['teardown'].duration
                                    if 'teardown' in item_attempt_reports else 0.
                                )
                        ) * 1000.
                    ),
                    'result': attempt_result,
                }
                item_attempts.append(attempt)

            if len(item_attempts) > 0:
                run_record: TestRunRecord = {
                    'filename': test_filename,
                    'name': list(test_name),
                    'attempts': item_attempts,
                }
                test_runs.append(run_record)

        request = {
            'start_time': _ts_to_rfc3339(self.start_time),
            'end_time': _ts_to_rfc3339(self.end_time),
            'test_runs': test_runs,
        }
        request.update(**({'branch': self.branch} if self.branch is not None else {}))
        request.update(**({'commit': self.commit} if self.commit is not None else {}))

        return cast(CreateTestSuiteRunRequest, request)

    # Allows us to override the exit code if all the failures are quarantined. We need this to be a
    # wrapper so that the default hook still gets invoked and prints the summary line with the test
    # category counts.
    @pytest.hookimpl(hookwrapper=True, tryfirst=True)
    def pytest_sessionfinish(
            self,
            session: pytest.Session,
            exitstatus: int,
    ) -> Generator[None, None, None]:
        self.end_time = time()
        self.logger.debug('called hook pytest_sessionfinish')
        yield
        self.logger.debug('resumed yielded hook pytest_sessionfinish')

        request = self._build_test_suite_run_request(session)

        if self.upload_results and len(request['test_runs']) > 0:
            try:
                run_summary = create_test_suite_run(
                    request=request,
                    test_suite_id=self.test_suite_id,
                    api_key=self.api_key,
                    base_url=self.base_url,
                    insecure_disable_tls_validation=self.insecure_disable_tls_validation,
                    logger=self.logger,
                )
            except Exception as e:
                pytest.exit('ERROR: Failed to report results to Unflakable: %s\n' % (repr(e)), 1)
            else:
                print(
                    'Unflakable report: %s' % (
                        build_test_suite_run_url(
                            self.test_suite_id,
                            run_summary['run_id'],
                            self.base_url,
                        )),
                )

        # We multiply by 2 here because each quarantined test is double-counted by the Session: once
        # for the quarantined report, and once for the fake report that's used for logging errors.
        if session.testsfailed > 0 and session.testsfailed == self.num_tests_quarantined * 2:
            pytest.exit('All failed tests are quarantined', 0)



"""Unflakable plugin for PyTest."""

#  Copyright (c) 2022 Developer Innovations, LLC

import argparse

import os
from typing import TYPE_CHECKING

import pytest
import logging

from ._git import get_current_git_commit, get_current_git_branch
from ._plugin import UnflakablePlugin, QuarantineMode

if TYPE_CHECKING:
    Config = pytest.Config
    Parser = pytest.Parser
else:
    Config = object
    Parser = object


def pytest_addoption(parser: Parser) -> None:
    group = parser.getgroup('Unflakable')
    group.addoption(
        '--api-key-path',
        type=str,
        dest='unflakable_api_key_path',
        help=(
            'Path to file containing Unflakable API key. If not specified, UNFLAKABLE_API_KEY '
            'environment variable must be set.'
        )
    )
    group.addoption(
        '--base-url',
        type=str,
        dest='unflakable_base_url',
        help=argparse.SUPPRESS,
    )
    group.addoption(
        '--branch',
        type=str,
        dest='unflakable_branch',
        help='name of version control branch (overrides Git auto-detection).',
    )
    group.addoption(
        '--commit',
        type=str,
        dest='unflakable_commit',
        help='name of version control revision (overrides Git auto-detection).',
    )
    group.addoption(
        '--enable-unflakable',
        action='store_true',
        dest='unflakable_enable',
        help='enable Unflakable plugin.'
    )
    group.addoption(
        '--failure-retries',
        type=int,
        dest='unflakable_failure_retries',
        default='2',
        help='maximum number of times to retry each failed test. (default: 2)'
    )
    # For local testing only!
    group.addoption(
        '--insecure-disable-tls-validation',
        action='store_true',
        dest='unflakable_insecure_disable_tls_validation',
        help=argparse.SUPPRESS,
    )
    group.addoption(
        '--no-git-auto-detect',
        action='store_const',
        const=True,
        dest='unflakable_no_git_auto_detect',
        help='do not attempt to auto-detect the current branch and commit hash from Git.',
    )
    group.addoption(
        '--no-upload-results',
        action='store_const',
        const=True,
        dest='unflakable_no_upload_results',
        help='do not report test results to Unflakable.',
    )
    group.addoption(
        '--quarantine-mode',
        type=str,
        dest='unflakable_quarantine_mode',
        default='ignore_failures',
        help=(
            'controls the behavior of quarantined tests. (default: ignore_failures)'
        ),
        metavar='<ignore_failures|no_quarantine|skip_tests>'
    )
    group.addoption(
        '--test-suite-id',
        type=str,
        dest='unflakable_suite_id',
        help='Unflakable test suite ID (required if --enable-unflakable is provided).'
    )
    group.addoption(
        '--unflakable-log-level',
        type=str,
        dest='unflakable_log_level',
        help='log level for Unflakable plugin',
    )


def pytest_configure(config: Config) -> None:
    logger = logging.getLogger('pytest_unflakable')
    if config.option.unflakable_log_level is not None:
        logging.basicConfig()
        logger.setLevel(level=config.option.unflakable_log_level)

    if not config.getoption('unflakable_enable', False):
        return

    if config.option.unflakable_quarantine_mode.upper() not in QuarantineMode.__members__:
        raise pytest.UsageError(
            f'Unrecognized quarantine mode `{config.option.unflakable_quarantine_mode}`')
    else:
        quarantine_mode = QuarantineMode.__members__[
            config.option.unflakable_quarantine_mode.upper()]

    # workerinput is set for xdist workers, but we only want to enable the plugin on the controller.
    is_xdist_worker = hasattr(config, 'workerinput')

    if config.getoption('unflakable_suite_id') is None:
        raise pytest.UsageError('missing required argument --test-suite-id')

    # pytest-xdist workers don't make API calls and amy not have the API key available.
    if is_xdist_worker:
        api_key = ''
    elif config.option.unflakable_api_key_path is not None:
        with open(config.option.unflakable_api_key_path, 'r') as api_key_file:
            api_key = api_key_file.read()
    elif 'UNFLAKABLE_API_KEY' in os.environ:
        api_key = os.environ['UNFLAKABLE_API_KEY']
    else:
        raise pytest.UsageError('missing required environment variable `UNFLAKABLE_API_KEY`')

    branch = config.option.unflakable_branch
    commit = config.option.unflakable_commit
    git_auto_detect = not config.getoption('unflakable_no_git_auto_detect', False)
    if git_auto_detect and not is_xdist_worker:
        if commit is None:
            commit = get_current_git_commit()
            logger.debug('auto-detected commit `%s`', commit)

        if branch is None and commit is not None:
            branch = get_current_git_branch(commit, logger)
            logger.debug('auto-detected branch `%s`', branch)

    if is_xdist_worker and 'unflakable_manifest' in config.workerinput:  # type: ignore
        worker_manifest = config.workerinput['unflakable_manifest']  # type: ignore
    else:
        worker_manifest = None

    config.pluginmanager.register(UnflakablePlugin(
        api_key=api_key,
        base_url=config.option.unflakable_base_url,
        branch=branch,
        commit=commit,
        failure_retries=config.option.unflakable_failure_retries,
        insecure_disable_tls_validation=config.getoption(
            'unflakable_insecure_disable_tls_validation', False),
        quarantine_mode=quarantine_mode,
        test_suite_id=config.option.unflakable_suite_id,
        upload_results=not is_xdist_worker and (
            not config.getoption('unflakable_no_upload_results', False)),
        logger=logger,
        worker_manifest=worker_manifest,
        is_xdist_worker=is_xdist_worker,
    ))
